#
# Copyright (c) 2003, 2004, 2005 Tama Communications Corporation
#
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without
# modifications, as long as this notice is preserved.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# User configuration variables:
#
# DEFINE_DIRECTORY	Default account related directory.
#			(default: nothing)
#
# ACCOUNT_DEFINITION	Account definition file.
#			(default: nothing)
#
# USER_SLIP_FILES	User slip files.
#			(default: main.slip)
#
# OPENING_TRANSACTIONS	Opening transaction file.
#			(default: ../[previous period]/forward-opening.trans)
#
# QUIET_MODE		Quiet mode.
#			(default: verbose mode)
#
ifndef DEFINE_DIRECTORY
DEFINE_DIRECTORY := $(shell bknconfig DEFINE_DIRECTORY)
endif

ifndef ACCOUNT_DEFINITION
ACCOUNT_DEFINITION := $(DEFINE_DIRECTORY)/account.def
endif

ifndef USER_SLIP_FILES
USER_SLIP_FILES := main.slip
endif

ifdef QUIET_MODE
V=
else
V=-v
endif
#
# If you would like to add new snapshots then you should add them here.
#
SNAP_FILES=	initial.snap opening.snap income.snap profit.snap asset.snap closing.snap
BALANCE_FILES=	$(patsubst %.snap,%.balance,$(SNAP_FILES))
TRANS_FILES=	opening.trans \
		opening-transfer.trans \
		user-transactions.trans \
		forward-closing.trans \
		forward-opening.trans \
		closing-profit.trans \
		closing-transfer.trans
SCM_FILES=	$(patsubst %.trans,%.scm,$(TRANS_FILES))

TOP_TARGETS=	all final test clean snap balance snap-files balance-files

.DELETE_ON_ERROR:

.PHONY: $(TOP_TARGETS)

all: final balance

#
# The 'forward-opening.trans' is to carry-over the result of this period
# to the next.
# The 'closing.balance' will be empty file; But this is important because
# it is an evidence of the correctness of the calculation.
# Other files are generated with the process by which these two files
# are made.
#
final: forward-opening.trans

#
# Carry-over from the previous period.
#
# Initially, BOKIN load opening transactions to setup the initial values.
# If you define OPENING_TRANSACTIONS then BOKIN use it.
# Otherwise, BOKIN uses "../(n - 1)/forward-opening.trans".
#
ifndef OPENING_TRANSACTIONS
  CUR_PERIOD := $(shell bknconfig period:id)
  ifeq ($(CUR_PERIOD),)
    CUR_PERIOD := $(shell basename `pwd`)
  endif
  ifeq ($(CUR_PERIOD),0)
    OPENING_TRANSACTIONS = /dev/null
  else
    PRE_PERIOD := $(shell expr $(CUR_PERIOD) - 1)
    OPENING_TRANSACTIONS = ../$(PRE_PERIOD)/forward-opening.trans
  endif
endif
ifdef GOBACK
$(OPENING_TRANSACTIONS):
	(cd ../$(PRE_PERIOD); $(MAKE) forward-opening.trans)
endif
# End of !OPENING_TRANSACTIONS

CLEAN_FILES=	$(SNAP_FILES) $(BALANCE_FILES) $(TRANS_FILES) $(SCM_FILES)

#
# Conversion rules.
#
# If you don't offer the following transactions then default transaction
# file will be used.
#
#	opening-transfer.trans
#	closing-profit.trans
#	closing-transfer.trans
#
.PRECIOUS: %.trans
vpath %.trans $(DEFINE_DIRECTORY)
%.scm: %.trans
	bknparse $(V) $< -o $@
%.balance: %.snap
	bknprint $(V) -b $< -o $@
%.pair: %.snap
	bknprint $(V) -p $< -o $@
#
# Generate rules for transaction files.
#
# Originally, because since *.trans files are intermediate, they are
# deleted later by make(1). But in BOKIN, all the intermediate files
# are leaved as it is to express the process of the calculation.
#
opening.trans: $(OPENING_TRANSACTIONS)
	bknsort $(V) $^ -o $@
user-transactions.trans: $(USER_SLIP_FILES)
	bknsort $(V) $^ -o $@
#
# Generate rules for each snapshot file.
# (Snapshots with underline are treated as part of finacial statement.)
#
# initial.snap
# v
# opening.snap	                    Asset, Liabilities, Initial Equity
# v
# income.snap	Expenses, Revenues, Asset, Liabilities, Initial Equity
# v             ~~~~~~~~  ~~~~~~~~
# profit.snap	Profit,             Asset, Liabilities, Initial Equity
# v             ~~~~~~ 
# asset.snap	Asset, Liabilities, Equity
# v             ~~~~~  ~~~~~~~~~~~  ~~~~~~
# closing.snap	(all in balance)
#
initial.snap: $(ACCOUNT_DEFINITION)
	@if [ "$^" = "" ]; then \
		echo "ERROR: ACCOUNT_DEFINITION variable not defined."; \
		exit 1; \
	fi
	bknapply $(V) $^ /dev/null -o $@
opening.snap: initial.snap opening.scm opening-transfer.scm
	bknapply $(V) $^ -o $@
income.snap: opening.snap user-transactions.scm
	bknapply $(V) $^ -o $@
profit.snap: income.snap closing-profit.scm
	bknapply $(V) $^ -o $@
asset.snap: profit.snap closing-transfer.scm
	bknapply $(V) $^ -o $@
closing.snap: asset.snap forward-closing.scm
	bknapply $(V) $^ -o $@
forward-opening.trans: asset.snap
	bkntrans $(V) -p $< -o $@
forward-closing.trans: asset.snap
	bkntrans $(V) -c $< -o $@

snap: $(SNAP_FILES)
snap-files:
	@echo $(SNAP_FILES)
balance: $(BALANCE_FILES)
balance-files:
	@echo $(BALANCE_FILES)
clean:
	-rm -f $(CLEAN_FILES) $(USER_CLEAN_FILES)
