;;
;; Copyright (c) 2003, 2004 Tama Communications Corporation
;;
;; This file is part of BOKIN.
;;
;; BOKIN is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; BOKIN is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software Foundation, Inc.
;; 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
;;
;; Account processing for BOKIN.

;;; Account definition

(define-module (bokin account)
:export (
	bkn:shortest-unique-path
	bkn:leaf?
	bkn:tree-ref
	bkn:node-ref
	bkn:load-account-file
	bkn:check-account-tree
	bkn:get-balance
	bkn:ancestor?
	bkn:add-pair!
	bkn:traverse-tree
	bkn:full-path
	bkn:get-account
	bkn:print-tree
	bkn:get-special-node
))

(use-modules ((ice-9 regex)
              :select (match:substring match:count match:suffix)))
(use-modules ((ice-9 rdelim)
              :select (read-line)))
(use-modules ((srfi srfi-13)
              :select (string-join)))
(use-modules (bokin error))
(use-modules (bokin config))

;;; The definition of data structure

;; Account tree.

(define TREE-ROOT 0)
(define TREE-INDEX 1)
(define TREE-CACHE 2)
(define TREE-LIST-SIZE 3)

;; Account node.

(define ACCT-USER-LEVEL 0)
(define ACCT-NAME 1)
(define ACCT-ALT-NAME 2)
(define ACCT-PAIR 3)
(define ACCT-TYPE 4)
(define ACCT-LEVEL 5)
(define ACCT-PARENT 6)
(define ACCT-CHILDREN 7)
(define ACCT-NODE-SIZE 8)

;;; The definition of account file.

(define number "[0-9]+")
(define value "[0-9]+|[0-9]+\\.[0-9]+")
(define name "[^ \t]+")
(define dot "\\.")
(define SEP "[ \t]+")
(define PAD "[ \t]*")
(define (extract arg)
  (string-append "(" arg ")"))

;; Meta record

(define pat-meta "^%(debit|credit|open|close)$")

;; Account

(define pat-def (string-append
	PAD
	(extract number)
	SEP
	(extract name)
	PAD
))

;; Pair

(define pat-pair (string-append
        PAD
	"\\("
	PAD
	(extract value)
	SEP
	dot
	SEP
	(extract value)
	PAD
	"\\)"
))

;; Account file input object.
;;
(define (account-file-object)
  (let ((port '())
        (cache '())
        (position '())
        (eof #f)
        (regex-blanks (make-regexp "^[ \t]*$"))
        (regex-meta (make-regexp pat-meta))
        (regex-def (make-regexp pat-def))
        (regex-pair (make-regexp pat-pair)))

    ; raw-read-line
    (define (raw-read-line)
      (let loop ((line (read-line port)))
        (if (eof-object? line)
          (begin
            (set! eof #t)
            '())
          ; NULL or comment line.
          (if (or (regexp-exec regex-blanks line)
                  (char=? bkn:comment-char (string-ref line 0)))
              (loop (read-line port))
              (let ((match (regexp-exec regex-meta line)))
                (if match
                    ; meta record
                    (let ((result (match:substring match 1)))
                      (cond ((string=? result "debit")
                             (set! position 'DEBIT))
                            ((string=? result "credit")
                             (set! position 'CREDIT))
                            ((string=? result "open")
                             (set! position 'OPEN))
                            ((string=? result "close")
                             (set! position 'CLOSE)))
                      (loop (read-line port)))
                    ; account record
                    line))))))

    ; next-account
    ;
    (define (next-account)
      (let ((line (raw-read-line)))
        (if (null? line)
            line
            (let ((level 0)
                  (name '())
                  (alt '())
                  (debit 0)
                  (credit 0)
                  (match (regexp-exec regex-def line)))
              (if (and match (= (match:count match) 3))
                  (begin
                    (set! level (string->number (match:substring match 1)))
                    (set! name (match:substring match 2))
                    ; Alternative name when 0 > n
                    (let ((ls (string-split name #\/)))
                      (if (= 2 (length ls))
                          (begin
                            (set! name (car ls))
                            (set! alt (cadr ls)))))
                    (let ((match-pair (regexp-exec regex-pair (match:suffix match))))
                      (if (and match-pair (= (match:count match-pair) 3))
                          (begin
                            (set! debit (string->number (match:substring match-pair 1)))
                            (set! credit (string->number (match:substring match-pair 2)))))
                      (list level name alt (cons debit credit) position)))
                    (bkn:error "illegal record.[~a]" line))))))
    (lambda (symbol . args)
      (case symbol
            ((OPEN)
              (if (not (pair? args))
                  (bkn:error "Method 'OPEN' needs an argument."))
                  (let ((file (car args)))
                    (if (port? file)
                        (set! port file)
                        (begin
                          (if (not (access? file R_OK))
                              (bkn:error "File '~a' not found." file))
                          (set! port (open-input-file (car args)))))
                    (set! position '())))
            ((PEEK)
              (if (null? cache)
                  (set! cache (next-account)))
              cache)
            ((READ)
              (if eof
                  '()
                  (let ((line (if (null? cache)
                                  (next-account)
                                  (let ((buf cache))
                                    (set! cache '())
                                    buf))))
                    line)))
            ((CLOSE)
              (close-input-port port))))))

;;; Loading account tree.

;; Load an account and the descendant.
;;
(define (make-node acct level acct-obj)
  (let ((children '())
        (node (make-vector ACCT-NODE-SIZE)))
    (do ((next (acct-obj 'PEEK) (acct-obj 'PEEK)))
        ; (car next) (car acct) is a user level number.
        ((or (null? next) (<= (car next) (car acct)))
         (vector-set! node ACCT-USER-LEVEL (list-ref acct 0))
         (vector-set! node ACCT-NAME (list-ref acct 1))
         (vector-set! node ACCT-ALT-NAME (list-ref acct 2))
         (vector-set! node ACCT-PAIR (list-ref acct 3))
         (vector-set! node ACCT-TYPE (list-ref acct 4))
         (vector-set! node ACCT-LEVEL level)
         (vector-set! node ACCT-PARENT '())
         (vector-set! node ACCT-CHILDREN (reverse children))
         node)
      (let ((child (make-node (acct-obj 'READ) (+ level 1) acct-obj)))
        (vector-set! child ACCT-PARENT node)
        (set! children (cons child children))))))


;; Make index of account tree.
;;
(define (make-index node-list)
  (let ((index (make-vector (count-node node-list)))
        (i 0))
    (let ((proc (lambda (node)
            (vector-set! index i node)
            (set! i (+ i 1)))))
      (bkn:traverse-tree proc node-list))
    (let ((proc (lambda (m1 m2)
            (string<? (vector-ref m1 ACCT-NAME) (vector-ref m2 ACCT-NAME)))))
      (set! index (sort! index proc))
      index)))

;; Get list of accounts which match to the name.
;;
(define (lookup-account tree name)

  ; compare procedure
  (define (compare name node)
    (let ((target (vector-ref node ACCT-NAME)))
      (cond ((string<? name target)
             'less)
            ((string>? name target)
             'greater)
            ((string=? name target)
             'match))))

  ; Execute binary search and return the index of one of the targets.
  (define (bsearch name array start end)
    (if (>= start end)
        -1
        (let* ((middle (quotient (+ start end) 2))
               (target (vector-ref array middle)))
          (case (compare name target)
            ((match)
             middle)
            ((less)
             (bsearch name array start middle))
            ((greater)
             (bsearch name array (+ middle 1) end))
            (else
             (bkn:error "bsearch: Illegal result."))))))

  (let ((array (list-ref tree TREE-INDEX))
        (limit (vector-length (list-ref tree TREE-INDEX))))
    (let ((ls '())
          (one (bsearch name array 0 limit)))
      (if (>= one 0)
          (let ((start (do ((i one (- i 1)))
                     ((or (< i 0)
                          (not (eq? 'match (compare name (vector-ref array i)))))
                      (+ i 1)))))
            (do ((i start (+ i 1)))
                ((or (>= i limit)
                     (not (eq? 'match (compare name (vector-ref array i))))))
              (set! ls (cons (vector-ref array i) ls)))))
      ls)))
        
;; Load account definition from file.
;;
;;	file	account definition file or snapshot
;;		both file name and input port are acceptable.
;;	result	account tree
;;
(define (bkn:load-account-file file)
  (let ((node-list '())
        (acct-obj (account-file-object)))
    (acct-obj 'OPEN file)
    (do ((acct (acct-obj 'READ) (acct-obj 'READ)))
        ((null? acct)
         (acct-obj 'CLOSE)
         (list node-list (make-index node-list) (make-hash-table 31)))
      (if (= (car acct) 1)
        (set! node-list (append! node-list (list (make-node acct 1 acct-obj))))
        (bkn:error "illegal level number.")))))

(define (bkn:check-account-tree tree)
  (if (not (and (list? tree) (= TREE-LIST-SIZE (length tree))))
      (bkn:error "Illegal account tree.")))

;;; The operation of account node.

; (define (bkn:chop pair)
;   (if (not (pair? args))
;       (bkn:error "chop: invalid argument."))
;   (let ((new (cons 0 0)))
;     (if (> (car pair) (cdr pair))
;         (set-car! new (- (car pair) (cdr pair)))
;         (set-cdr! new (- (cdr pair) (car pair))))
;     new))

(define (bkn:get-balance node)
  (let ((pair (vector-ref node ACCT-PAIR)))
    (if (eq? (vector-ref node ACCT-TYPE) 'DEBIT)
        (- (car pair) (cdr pair))
        (- (cdr pair) (car pair)))))


;; Whether or not self is one of the ancestor of target node.
;;
(define (bkn:ancestor? tree self target)
  (let loop ((node target))
    (if (null? node)
        #f
        (if (eq? self node)
            #t
            (loop (parent node))))))

(define (bkn:tree-ref tree sym)
  (list-ref tree (cond
                    ((eq? sym 'root) TREE-ROOT)
                    ((eq? sym 'index) TREE-INDEX)
                    ((eq? sym 'cache) TREE-CACHE)
                    (else (bkn:error "Such member not found.")))))

(define (bkn:node-ref node sym)
  (vector-ref node (cond
                    ((eq? sym 'user-level) ACCT-USER-LEVEL)
                    ((eq? sym 'name) ACCT-NAME)
                    ((eq? sym 'alt-name) ACCT-ALT-NAME)
                    ((eq? sym 'pair) ACCT-PAIR)
                    ((eq? sym 'type) ACCT-TYPE)
                    ((eq? sym 'level) ACCT-LEVEL)
                    ((eq? sym 'parent) ACCT-PARENT)
                    ((eq? sym 'children) ACCT-CHILDREN)
                    (else (bkn:error "Such member not found.")))))

; Dump the given node for debugging.
(define (bkn:node-dump node)
  (display ", level=")
  (display (vector-ref node ACCT-LEVEL))
  (display "user-level=")
  (display (vector-ref node ACCT-USER-LEVEL))
  (display ", type=")
  (display (vector-ref node ACCT-TYPE))
  (display ", name=")
  (display (vector-ref node ACCT-NAME))
  (display ", alt-name=")
  (display (vector-ref node ACCT-ALT-NAME))
  (display ", pair=")
  (display (vector-ref node ACCT-PAIR))
  (newline))

(define (children node)
  (vector-ref node ACCT-CHILDREN))

(define (bkn:leaf? node)
  (null? (children node)))

(define (parent node)
  (vector-ref node ACCT-PARENT))

(define (has-parent? node)
  (not (null? (parent node))))

(define (bkn:null? pair)
  (if (pair? pair)
      (and (= (car pair) 0) (= (cdr pair) 0))
      (bkn:error "Invalid pair.")))

;; Enbalance all leaf node in account tree.
;;
(define (enbalance node-list)
  (define (balance-pair node)
    (let* ((pair (vector-ref node ACCT-PAIR))
           (debit (car pair))
           (credit (cdr pair)))
      (if (> debit credit)
          (cons 0 (- debit credit))
          (cons (- credit debit) 0))))
  (let ((proc (lambda (node)
          (if (bkn:leaf? node)
              (bkn:add-pair! node (balance-pair node))))))
    (bkn:traverse-tree proc node-list)))

;; Add pair value to the pair value of the node.
;;
;;	node	specified node
;;	value	pair value
;;
(define (bkn:add-pair! node value)
  (let ((need-exchange (or (and (< (car value) 0) (= (cdr value) 0))
                           (and (< (cdr value) 0) (= (car value) 0)))))
  ;
  ; (0 . -999) => (999 . 0)
  ; (-999 . 0) => (0 . 999)
  ;
  (if need-exchange
      (let ((tmp (car value)))
        (set-car! value (- 0 (cdr value)))
        (set-cdr! value (- 0 tmp))))
  (if (bkn:leaf? node)
      ;
      ; If it is a leaf account, you can execute all calculations.
      ;
      (let loop ((pair (vector-ref node ACCT-PAIR)))
        (set-car! pair (+ (car pair) (car value)))
        (set-cdr! pair (+ (cdr pair) (cdr value)))
        (if (has-parent? node)
            (begin
              (set! node (parent node))
              (loop (vector-ref node ACCT-PAIR)))))
      ;
      ; If it isn't a leaf account, you can execute add-pair!
      ; when the calculation brings balance 0.
      ;
      (let* ((pair (vector-ref node ACCT-PAIR))
             (debit (+ (car pair) (car value)))
             (credit (+ (cdr pair) (cdr value))))
        (if (= debit credit)
            (begin
              ; Enbalance all leaf accounts.
              (enbalance (children node)))
            (bkn:error "Must be a leaf node.[~a]" (vector-ref node ACCT-NAME)))))))

;;; Account tree opration.

;; Traverse an account tree and execute specified procedure for each account.
;;
(define (bkn:traverse-tree proc node-list)
  (for-each
    (lambda (node)
      (proc node)
      (bkn:traverse-tree proc (children node)))
    node-list))

;; Count nodes in an account tree.
;;
(define (count-node node-list)
  (let ((count 0))
    (let ((proc (lambda (node)
            (set! count (+ count 1)))))
      (bkn:traverse-tree proc node-list)
      count)))

;; Get an special account node.
;;
;; This procedure is used for getting open or close account.
;;
(define (bkn:get-special-node tree type)
  (let ((result #f))
    (let ((proc (lambda (node)
            (if (eq? type (vector-ref node ACCT-TYPE))
                (set! result node)))))
      (bkn:traverse-tree proc (list-ref tree TREE-ROOT))
      result)))

;; Calculate the full path of the specified node.
;;
(define (bkn:full-path node)
  (let ((parent (vector-ref node ACCT-PARENT)))
    (if (null? parent)
        (string-append "/" (vector-ref node ACCT-NAME))
        (string-append (bkn:full-path parent) "/" (vector-ref node ACCT-NAME)))))

;; Calculate the shortest path of the specified node.
;;
(define (bkn:shortest-unique-path tree node)
  (let ((path (string-split (bkn:full-path node) #\/)))
    (do ((ls (cdr path) (cdr ls)))
        ((or (null? ls)
             (eq? 'DUPLICATED (bkn:get-account tree (string-join ls "/"))))
         (string-join path "/"))
      (set! path ls))))

;; Try matching test for a node.
;;
;; The path is ordered from leaf to root.
;;
(define (match-account node path only-root)
  (let ((u (if (null? path)
               '()
               (car path)))		; Pick up the first unit of path.
        (rest (if (null? path)
                  '()
                  (cdr path))))		; The rest of path.
    (cond

      ; 1. If it is the first unit of the path.
      ;
      ((null? rest)
        (cond
          ((null? node)
           #t)	; It means match to the root.
          ((and (string? u) (string=? (vector-ref node ACCT-NAME) u))
           #t)
          (else
           #f)))

      ; 2. From now on, the middle unit of the path.
      ;
      ((null? node)
       #f)

      ; 3. From now on, before reaching the root of the path.
      ;
      ((not (string=? u (vector-ref node ACCT-NAME)))
        #f)
      (else
        (match-account (parent node) rest only-root)))))
           
;; Get an account node by name.
;;
;;	tree		account tree
;;	name		account name
;;	NOTFOUND	The node is not found.
;;	DUPLICATED	The node is duplicated.
;;
;;	tree == (root-account-list index cache)
;;
(define (bkn:get-account tree name)
  ; Pick up the last item.
  (define (lastof ls)
    (do ((p ls (cdr p)))
        ((or (null? p) (null? (cdr p)))
         (car p))))

  (let* ((hash (list-ref tree TREE-CACHE))
         (sym (string->symbol name))
         (node (hashq-ref hash sym)))
    ; If exists in cache then return the value.
    (if (vector? node)
        node
        ; If not exist in cache then lookup it.
        (let ((path (string-split name #\/))
              (only-root #f))
          ; Account name start with '/'.
          (if (string=? "" (car path))
              (begin
                (set! only-root #t)
                (set! path (cdr path))))
          ; Check illegal path name.
          (for-each
            (lambda (unit)
              (if (string=? "" unit)
                  (bkn:error "Illegal account path name.")))
           path)
          (if (= 0 (length path))
              (bkn:error "Illegal account path name."))

          ; Get candidates list from the tail component and check whether
          ; or not match to the path for each candidate.
          ; Both no match and duplicate match are assumed error.
          ;
          (let ((result 'NOTFOUND))
            (let loop ((cand (lookup-account tree (lastof path))))
              (if (not (null? cand))
                  (if (match-account (car cand) (reverse path) only-root)
                      (begin
                        (if (eq? result 'NOTFOUND)
                            (begin
                              (set! result (car cand))
                              (loop (cdr cand)))
                            (set! result 'DUPLICATED)))
                      (begin
                        (loop (cdr cand))))))
            (if (vector? result)
                (hashq-create-handle! hash sym result))
            result)))))

;; Print an account tree.
;;
;;	tree == (root-account-list index cache)
;;
(define (bkn:print-tree oport tree . options)
  (let ((position '())
        (opt:level 0)
        (opt:alternative #f)
        (opt:balance #f)
        (opt:pair #f)
        (opt:snap #f))

    ; Parse options.

    (if (null? options)
        (set! opt:snap #t)
        (for-each
          (lambda (a)
            (if (integer? a)
                (set! opt:level a)
                (case a
                  ((alternative)
                   (set! opt:alternative #t))
                  ((balance)
                   (set! opt:balance #t))
                  ((pair)
                   (set! opt:pair #t))
                  ((snap)
                   (set! opt:snap #t)))))
          options))

    ; Print a node.

    (let ((proc (lambda (node)
            (if (not (eq? position (vector-ref node ACCT-TYPE)))
                (begin
                  (set! position (vector-ref node ACCT-TYPE))
                  (if opt:snap
                      (begin
                        (case position
                          ((DEBIT)
                           (display "%debit" oport))
                          ((CREDIT)
                           (display "%credit" oport))
                          ((OPEN)
                           (display "%open" oport))
                          ((CLOSE)
                           (display "%close" oport))
                          (else
                           (bkn:error "unknown account type.")))
                        (newline oport)))))
            (let ((pair (vector-ref node ACCT-PAIR))
                  (balance (bkn:get-balance node))
                  (n (vector-ref node ACCT-LEVEL)))
              (if (and (or (= opt:level 0)
                           (<= n opt:level))
                       (or opt:snap
                           (and opt:balance (not (= balance 0)))
                           (and opt:pair (not (bkn:null? pair)))))
                  (begin
                    ; Level number is indented by the logical level.
                    (if (> n 1)
                        (display (make-string (* (- n 1) 3) #\space) oport))
                    (display #\0 oport)
                    (display (vector-ref node ACCT-USER-LEVEL) oport)
                    (display #\space oport)
                    ;
                    ; Some accounts have alternative name. For example,
                    ; 'ordinary-income' is called 'ordinary-loss' when
                    ; the balance < 0.
                    ;
                    (let ((name (vector-ref node ACCT-NAME))
                          (alt-name (vector-ref node ACCT-ALT-NAME)))
                      (cond (opt:snap     ; Print name as is.
                             (display name oport)
                             (if (not (null? alt-name))
                                 (begin
                                   (display "/" oport)
                                   (display alt-name oport))))
                            (else         ; Print name alternatively.
                             (if (and (< balance 0) opt:alternative (not (null? alt-name)))
                                 (display alt-name oport)
                                 (display name oport)))))
                    (display #\space oport)
                    (if opt:snap
                        (display pair oport)
                        (begin
                          (if opt:balance
                              (display balance oport))
                          (if (and opt:balance opt:pair)
                              (display " " oport))
                          (if opt:pair
                              (display pair oport))))
                    (newline oport)))))))
      (bkn:traverse-tree proc (list-ref tree TREE-ROOT)))))
