;;
;; Copyright (c) 2003, 2004 Tama Communications Corporation
;;
;; This file is part of BOKIN.
;;
;; BOKIN is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; BOKIN is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software Foundation, Inc.
;; 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
;;
;; Parser for transaction file for BOKIN

(define-module (bokin parse)
:export (
	bkn:parse-transaction
))

(use-modules ((ice-9 rdelim)
              :select (read-line read-delimited)))
(use-modules ((srfi srfi-13)
              :select (string-join reverse-list->string)))
(use-modules (bokin date))
(use-modules (bokin account))
(use-modules (bokin error))

;; Transaction format
;;
;;	transaction := date separator debit separator credit separator note \n
;;	separator := [ \t]+
;;	date := yyyymmdd[/serial]
;;	debit := account-name '/' value ( ',' account-name '/' value )* 
;;	credit := account-name '/' value ( ',' account-name '/' value )* 
;;	note := [^\n]*
;;

;;;; Character definitions.

(define BEGIN-BRACE #\()
(define PATH-SEP #\/)
(define MINUS #\-)
(define COMMA #\,)

;;; Input expression.

;; Read numeric value.
;;
;;	port	input port
;;	result	scheme value of numeric
;;
;;	format: [0-9]+|[0-9]\\.[0-9]+
;;
(define (read-numeric-value port)
  (let ((dot #\.))
    (string->number (list->string (reverse
      (let loop ((val '())(c (read-char port)))
        (if (eof-object? c)
            val
            (if (char-numeric? c)
                (loop (cons c val) (read-char port))
                (if (and dot (char=? c dot))
                    (begin
                      (set! dot #f)
                      (loop (cons c val) (read-char port)))
                    (begin
                      (unread-char c port)
                      val))))))))))
  
;; Read numeric value or S-expression.
;;
;;	port	input port
;;	result	scheme value of numeric or S-expression
;;
(define (read-expr port)
  (let ((c (peek-char port)))
    (if (eof-object? c)
        (bkn:error "Unexpected end of data.")
        (cond
          ; S-expression
          ((char=? BEGIN-BRACE c)
           (read port))
          ; Minus integer
          ((char=? MINUS c)
           (read-char)
           (if (char-numeric? (peek-char port))
               (- 0 (read-numeric-value port))
               (bkn:error "Not integer.")))
          ; Integer
          ((char-numeric? c)
           (read-numeric-value port))
          (else
           (bkn:error "Neither numeric value nor S-expression."))))))
      
;; Read pair
;;
;;	port	input port
;;	result	scheme value of pair
;;
(define (read-pair position port)
  (let ((root (char=? PATH-SEP (peek-char port)))
        (name '())
        (expr 0)
        (unit '())
        (slash #f))
    ; The name which start with '/' is an absolute path name.
    (if root
        (read-char port))
    (let loop ((c (peek-char port)))
      (cond
        ; Error.
        ((or (eof-object? c) (char-whitespace? c) (char=? COMMA c))
         (if (null? unit)
             (bkn:error "Unexpected end of data."))
         (set! name (cons (reverse-list->string unit) name))
         (set! unit '()))
        ; Separator
        ((or (char=? PATH-SEP c) (char-whitespace? c))
         (if (null? unit)
             (bkn:error "Illegal account name."))
         (set! slash (char=? PATH-SEP (read-char port)))
         (set! name (cons (reverse-list->string unit) name))
         (set! unit '())
         (if (char=? PATH-SEP c)
             (loop (peek-char port))))	; Go to next value processing.
        ; Input value.
        ((and slash (or (char=? c BEGIN-BRACE) (char-numeric? c)))
         (if (null? name)
             (bkn:error "Name not found."))
         (set! slash #f)
         (set! expr (read-expr port)))
        (else
         (set! unit (cons (read-char port) unit))
         (set! slash #f)
         (loop (peek-char port)))))
    (let ((path-sep-string (make-string 1 PATH-SEP)))
      (list (string-append (if root path-sep-string "")
                           (string-join (reverse name) path-sep-string))
            (if (eq? position 'CREDIT)
                (cons 0 expr)
                (cons expr 0))))))

;; Read date.
;;
;;	port	input port
;;	result	scheme value of date
;;
(define (read-date port)
  ; Read only date part of a transaction.
  (let* ((pair (read-delimited "\t " port 'split))
         (line (car pair))
         (delim (cdr pair)))
    (if (or (eof-object? line) (eof-object? delim))
        (bkn:error "Unexpected end of record.")
        (let ((date (bkn:string->date line)))
          (if date
              date
              (bkn:error "Illegal date format."))))))

;; Skip blanks and return the first character which is not blank.
;; Since this procedure set file pointer to this position, you can
;; re-read this character.
;;
(define (skip-blanks port)
  (do ((c (peek-char port) (peek-char port)))
      ((or (eof-object? c) (not (or (char=? #\space c) (char=? #\tab c))))
       c)
    (read-char port)))

;; Parse text image transaction
;;
;;	line	text image transaction
;;	result	scheme value of transaction
;;
(define (bkn:parse-transaction text)
  (if (= (string-length text) 0)
      (bkn:error "Transaction length 0."))
  ; Setup error procedure.
  (bkn:set-current-line text)
  (call-with-input-string text
    (lambda (port)
      (let ((date (read-date port))
            (pair-list '()))
        (if (not date)
            (bkn:error "Transaction has no date."))
        (if (eof-object? (skip-blanks port))
            (bkn:error "Transaction has no debit."))
        (let loop ((pair (read-pair 'DEBIT port)))
          (set! pair-list (append! pair-list (list pair)))
          (let ((c (skip-blanks port)))
            (if (and (not (eof-object? c)) (char=? COMMA c))
              (begin
                (read-char port)
                (skip-blanks port)
                (loop (read-pair 'DEBIT port))))))
        (if (eof-object? (skip-blanks port))
            (bkn:error "Transaction has no credit."))
        (let loop ((pair (read-pair 'CREDIT port)))
          (set! pair-list (append! pair-list (list pair)))
          (let ((c (skip-blanks port)))
            (if (and (not (eof-object? c)) (char=? COMMA c))
              (begin
                (read-char port)
                (skip-blanks port)
                (loop (read-pair 'CREDIT port))))))
        (let ((comment (read-line port)))
          (if (eof-object? comment)
              (set! comment (make-string 0)))
          (bkn:unset-current-line)
          (list date pair-list))))))
