;;
;; Copyright (c) 2003, 2004 Tama Communications Corporation
;;
;; This file is part of BOKIN.
;;
;; BOKIN is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; BOKIN is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software Foundation, Inc.
;; 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
;;
;; Transaction processing for BOKIN

(define-module (bokin transaction)
:export (
	bkn:transaction-apply
	bkn:transaction-prepare
	bkn:transaction-print
	bkn:transaction-execute
	TRANS-NAME
	TRANS-PAIR
	TRANS-NODE
))

(use-modules (bokin directory))
(use-modules (bokin error))
(use-modules (bokin account))
(use-modules (bokin date))

;;; Transaction procedures.

(define current-tree #f)
(define current-date #f)
(define current-node #f)

(define TRANS-NAME 0)
(define TRANS-PAIR 1)
(define TRANS-NODE 2)

(define user-custom-file "custom.scm")

;; Read S-transactions from iport and apply them to given tree.
;;
(define (bkn:transaction-apply tree iport verbose)
  (do ((S-expr (read iport) (read iport)))
      ((eof-object? S-expr))
    (if verbose
        (begin
          (display S-expr (current-error-port))
          (newline (current-error-port))))
    (bkn:transaction-prepare tree S-expr)
    (apply bkn:transaction-execute tree S-expr)))

;; Convert account name into node.
;;
(define (bkn:transaction-prepare tree S-expr)
  (for-each
    (lambda (item)
      (let ((node (bkn:get-account tree (list-ref item TRANS-NAME))))
        (if (vector? node)
            (append! item (list node))
            (bkn:error "Account '~a' not found or duplicated." (list-ref item TRANS-NAME)))))
    (cadr S-expr)))

;; Print transaction.
;;
(define (bkn:transaction-print S-expr)
  (define (debit item)
    (car (list-ref item TRANS-PAIR)))

  ; Print date
  (display (bkn:date->string (list-ref S-expr 0)))
  (display #\tab)

  ; Print comment if any.
  (let ((separator ""))
    (for-each
      (lambda (item)
        (let ((pair (list-ref item TRANS-PAIR)))
          (if (> (car pair) 0)
              (begin
                (display separator)
                (display (list-ref item TRANS-NAME))
                (let ((value (car pair)))
                  (if (> value 0)
                      (begin
                        (display #\/)
                        (display value))))
                (set! separator ",")))))
      (list-ref S-expr 1))
    (display #\tab)
    (set! separator "")
    (for-each
      (lambda (item)
        (let ((pair (list-ref item TRANS-PAIR)))
          (if (> (cdr pair) 0)
              (begin
                (display separator)
                (display (list-ref item TRANS-NAME))
                (let ((value (cdr pair)))
                  (if (> value 0)
                      (begin
                        (display #\/)
                        (display value))))
                (set! separator ",")))))
      (list-ref S-expr 1))))

;; Transaction processing.
;;
;;	tree	Account tree
;;	date	Transaction date(year month day [serial])
;;	list	(Account name (debit . credit))
;;	comment	(comment)
;;
(define (bkn:transaction-execute tree date pairs)
  (if (< (transaction-check&completion tree date pairs) 0)
      (bkn:error "transaction not balanced.\n~a ~a" date pairs)
      (for-each
        (lambda (item)
          (bkn:add-pair! (list-ref item TRANS-NODE) (list-ref item TRANS-PAIR)))
        pairs)))

;; Consistency check and completion.
;;
;;	result	0: ok, 1: ok and opening account, -1: error
;;
(define (transaction-check&completion tree date pairs)
  (let ((diff 0)
        (sum:debit 0)
        (sum:credit 0)
        (null:pair '())
        (result 0))

    ; Set current account tree to current-tree during transaction processing.
    (set! current-tree tree)
    (set! current-date date)

    ; Check whether or not it is balanced.
    ;
    ;	item: ("account" (1000 . 0))
    (for-each
      (lambda (item)
        (let ((node (list-ref item TRANS-NODE))
              (pair (list-ref item TRANS-PAIR)))

          ;; Start of eval.

          ; Scheme expressions in a transaction is evaluated here.
          (set! current-node node)
          ; If the element of the pair is not number then it's scheme
          ; expression. It should be evaluated into number.
          (if (not (number? (car pair)))
              (set-car! pair (local-eval (car pair) (the-environment))))
          (if (not (number? (cdr pair)))
              (set-cdr! pair (local-eval (cdr pair) (the-environment))))
          ;; End of eval.

          (set! current-node #f)
          ;(if (inexact? (car pair))
          ;    (bkn:error "Inexact number is not acceptable. (~a/~a)." (bkn:node-ref node 'name) (car pair)))
          ;(if (inexact? (cdr pair))
          ;    (bkn:error "Inexact number is not acceptable. (~a/~a)." (bkn:node-ref node 'name) (cdr pair)))
          (if (eq? (bkn:node-ref node 'type) 'OPEN)
              (set! result 1))
          (if (= (+ (car pair) (cdr pair)) 0)
              (set! null:pair pair)
              (begin
                (set! sum:debit (+ sum:debit (car pair)))
	        (set! sum:credit (+ sum:credit (cdr pair)))))))
      pairs)
    (set! diff (- sum:debit sum:credit))
    (set! current-tree #f)
    (set! current-date #f)

    ; Auto completion if possible.
    (if (not (= diff 0))
        (if (null? null:pair)
            (set! result -1)
            (if (< diff 0)
                (set-car! null:pair (abs diff))
                (set-cdr! null:pair (abs diff)))))
    result))

;; Load user custom code.
;;
(define (bkn:load-custom . option)
  (let* ((verbose (if (pair? option) (car option) #f))
         (customlist (cdr (bkn:make-file-list user-custom-file verbose))))
    (for-each
      (lambda (path)
        (if verbose
            (bkn:warning "Loading ~s." path))
        (primitive-load path))
      customlist)))

(bkn:load-custom)

;;; Builtin procedures.

(define (balance . arg)
  (if (or (not current-tree) (not current-node)) 
      (bkn:error "Out of transaction."))
  (let* ((name (if (pair? arg)
                   (car arg)
                   #f))
         (node (if name
                   (bkn:get-account current-tree name)
                   current-node)))
    (cond
     ((eq? node 'NOTFOUND)
      (bkn:error "Account '~a' not found." name))
     ((eq? node 'DUPLICATE)
      (bkn:error "Account '~a' duplicated." name)))
    (bkn:get-balance node)))

(define (divide . args)
  (if (< (length args) 2)
      (bkn:error "Procedure divide needs two or more arguments."))
  (inexact->exact (truncate (apply / args))))

(define (multiply . args)
  (if (< (length args) 2)
      (bkn:error "Procedure multiply needs two or more arguments."))
  (inexact->exact (truncate (apply * args))))
